#include "gdsclient_agent.h"
#include "gds_command.h"
#include "gdsflight.h"
#include "gdsdatabase_flights.h"
#include "gdsdatabase_pnrs.h"
#include "gdsdatabase_hotels.h"
#include "gdsclientsbase.h"
#include "exception.h"

GDSClientAgent::GDSClientAgent( GDSClientConnection * connection, const QString& login )
: GDSClient( connection )
, m_login( login )
, m_state( VOID )
{
}

GDSClientAgent::~GDSClientAgent()
{
}

QString GDSClientAgent::generateID()
{
	const char syms[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
	QString result;
	for (int i = 0; i < 6; i++)
	{
		result += QString( syms[ qrand() % 36 ] );
	}
	return result;
}

bool GDSClientAgent::processContent( const QByteArray & data )
{
	m_log.proto( "Incoming content:", data );
	bool result = true;
	GDSCommand* cmd = GDSCommandCreator::create( data );
	try
	{
		if ( dynamic_cast<GDSCommandJump *>( cmd ) )
		{
			processJumpIn( cmd );
		}
		else if ( dynamic_cast<GDSCommandAvail *>( cmd ) )
		{
			processAvail( cmd );
		}
		else if ( dynamic_cast<GDSCommandHelp *>( cmd ) )
		{
			processHelp( cmd );
		}
		else if ( dynamic_cast<GDSCommandSellFlight *>( cmd ) )
		{
			processSellFlight( cmd );
		}
		else if ( dynamic_cast<GDSCommandSellHotel *>( cmd ) )
		{
			processSellHotel( cmd );
		}
		else if ( dynamic_cast<GDSCommandEndSell *>( cmd ) )
		{
			processEndSell( cmd );
		}
		else if ( dynamic_cast<GDSCommandAddPassenger *>( cmd ) )
		{
			processAddPassenger( cmd );
		}
		else if ( dynamic_cast<GDSCommandRetrieve *>( cmd ) )
		{
			processRetrievePNR( cmd );
		}
		else if ( dynamic_cast<GDSCommandDeletePNR *>( cmd ) )
		{
			processDeletePNR( cmd );
		}
		else if ( dynamic_cast<GDSCommandChangePassword *>( cmd ) )
		{
			processChangePassword( cmd );
		}
		else if ( dynamic_cast<GDSCommandHotelsList *>( cmd ) )
		{
			processHotelsList( cmd );
		}
		else if ( dynamic_cast<GDSCommandAbortSell *>( cmd ) )
		{
			processAbortSell( cmd );
		}
		else
		{
			result = false;
		}
	}
	catch (StateError& )
	{
		sendContent( "Wrong state" );
	}
	catch (DatabaseError& )
	{
		QFile f( GDSPnrDatabase::filename );
		f.open( QIODevice::ReadOnly );
		sendContent( f.readAll() );
		f.close();
	}
	catch (InternalError& )
	{
		sendContent( "You are doing it wrong" );
	}
	delete cmd;
	return result;
};

void GDSClientAgent::processJumpIn( GDSCommand* )
{
	stateCheck( m_state != VOID );
	sendContent( "** Welcome to SimpleDS **" );
	m_state = NORMAL;
}

void GDSClientAgent::processAvail( GDSCommand* cmd )
{
	stateCheck( m_state != NORMAL );
	GDSCommandAvail *c = dynamic_cast<GDSCommandAvail *>( cmd );
	m_availResults = GDSFlightDatabase::instance()->getByODD( c->origin(), c->destination(), c->date() );
	sendContent( "** SimpleDS Availability - AN ** " + c->origin() + c->date().toStringDate() );
	int i = 1;
	foreach( long id, m_availResults )
	{
		GDSFlight *f = GDSFlightDatabase::instance()->get(id);
		sendContent( QString( "%1 %2" ).arg( i ).arg( f->toPrettyString() ) );
		i++;
		delete f;
	}
	sendContent( QString() );
}

void GDSClientAgent::processHotelsList( GDSCommand* cmd )
{
	stateCheck( m_state != NORMAL );
	GDSCommandHotelsList *c = dynamic_cast<GDSCommandHotelsList *>( cmd );
	m_hotelsResults = GDSHotelDatabase::instance()->getByCD( c->city(), c->date() );
	sendContent( "** SimpleDS Hotels Availability - HA ** " + c->city() + c->date().toStringDate() );
	int i = 1;
	foreach( long h, m_hotelsResults )
	{
		GDSHotel *f = GDSHotelDatabase::instance()->get(h);
		sendContent( QString( "%1 %2" ).arg( i ).arg( f->toPrettyString() ) );
		i++;
		delete f;
	}
	sendContent( QString() );
}

void GDSClientAgent::processHelp( GDSCommand* cmd )
{
	stateCheck( m_state == VOID );
	GDSCommandHelp *c = dynamic_cast< GDSCommandHelp* >( cmd );
	QString topic = c->topic();
	sendContent( GDSCommandCreator::help( topic ) );
}

void GDSClientAgent::processSellFlight( GDSCommand* cmd )
{
	stateCheck( m_state != NORMAL );
	GDSCommandSellFlight *c = dynamic_cast< GDSCommandSellFlight* >( cmd );
	int line = 0;
	if ( c->isLong() )
	{
		m_availResults = GDSFlightDatabase::instance()->getByTemplate( GDSFlight( c ) );
	}
	else
	{
		line = c->line();
	}

	if ( line < 0 || m_availResults.length() <= line )
	{
		sendContent( "Not found" );
		return;
	}
	m_sellingAvia = true;


	m_currentFlightId = m_availResults[ line ];
	GDSFlight *flight = GDSFlightDatabase::instance()->get(m_currentFlightId);
	m_wantedClass = c->sclass();
	flight->setClass( m_wantedClass );
	m_wantedPNRs = c->number();
	if ( flight->getAvailableSeats() < m_wantedPNRs )
	{
		sendContent( QString ( "Not enough seats of class '%1' are available on this flight" ).arg( c->sclass() ) );
		return;
	}
	flight->decAvailableSeats( m_wantedPNRs );
	GDSFlightDatabase::instance()->update( m_currentFlightId, flight );
	m_currentPNRs.clear();
	sendContent( QString( "%1 %2" ).arg( QString::number( line + 1 ).rightJustified( 4 ) ).arg( flight->toShortPrettyString() ) );
	m_state = SELL;
	delete flight;
}

void GDSClientAgent::processSellHotel( GDSCommand* cmd )
{
	stateCheck( m_state != NORMAL );
	GDSCommandSellHotel *c = dynamic_cast< GDSCommandSellHotel* >( cmd );
	int line = c->line();
	if ( line < 0 || m_hotelsResults.length() <= line )
	{
		sendContent( "Not found" );
		return;
	}
	m_sellingAvia = false;
	m_currentHotelId = m_hotelsResults[ line ];

	GDSHotel *hotel = GDSHotelDatabase::instance()->get(m_currentHotelId);
	
	m_wantedClass = c->sclass();
	hotel->setClass( m_wantedClass );
	m_wantedPNRs = c->number();
	if ( hotel->getAvailableRooms() < m_wantedPNRs )
	{
		sendContent( QString ( "Not enough rooms of class '%1' are available in this hotel" ).arg( c->sclass() ) );
		return;
	}
	m_state = SELL;
	hotel->decAvailableRooms( m_wantedPNRs );
	GDSHotelDatabase::instance()->update( m_currentHotelId, hotel );
	m_currentPNRs.clear();
	sendContent( QString( "%1 %2" ).arg( QString::number( line + 1 ).rightJustified( 4 ) ).arg( hotel->toShortPrettyString() ) );
	delete hotel;
}

void GDSClientAgent::processEndSell( GDSCommand* )
{
	stateCheck( m_state != SELL );
	if ( m_currentPNRs.length() < m_wantedPNRs )
	{
		sendContent( "Not all of the passengers are filled in yet" );
		return;
	}
	QString uniqID = generateID();
	foreach( GDSPnr pnr, m_currentPNRs )
	{
		pnr.setUniqID( uniqID );
		pnr.setClass( m_wantedClass );
		GDSPnrDatabase::instance()->add( pnr );
	}
	sendContent( "END OF TRANSACTION COMPLETE - " + uniqID );
	m_state = NORMAL;
}

void GDSClientAgent::processAbortSell( GDSCommand* )
{
	stateCheck( m_state != SELL );
	if ( m_sellingAvia ) {
		GDSFlight *flight = GDSFlightDatabase::instance()->get(m_currentFlightId);
		flight->incAvailableSeats( m_wantedPNRs );
		GDSFlightDatabase::instance()->update(m_currentFlightId,flight);
		delete flight;
	}
	else {
		GDSHotel *hotel = GDSHotelDatabase::instance()->get(m_currentHotelId);
		hotel->incAvailableRooms( m_wantedPNRs );
		GDSHotelDatabase::instance()->update(m_currentHotelId,hotel);
		delete hotel;
	}
	sendContent( "SELL TRANSACTION ABORTED" );
	m_state = NORMAL;
}

void GDSClientAgent::processAddPassenger( GDSCommand* cmd )
{
	stateCheck( m_state != SELL );
	GDSCommandAddPassenger *c = dynamic_cast< GDSCommandAddPassenger* >( cmd );
	if ( m_currentPNRs.length() >= m_wantedPNRs )
	{
		sendContent( "Can't add more passengers" );
		return;
	}
	if ( m_sellingAvia )
		m_currentPNRs.append( GDSPnr( m_currentFlightId, c, m_login, m_sellingAvia ) );
	else
		m_currentPNRs.append( GDSPnr( m_currentHotelId, c, m_login, m_sellingAvia ) );
	sendContent( QString( "%1.%2" ).arg( m_currentPNRs.length() ).arg( m_currentPNRs.last().toShortPrettyString() ) );
}

void GDSClientAgent::processRetrievePNR( GDSCommand* cmd )
{
	stateCheck( m_state != NORMAL );
	GDSCommandRetrieve *c = dynamic_cast<GDSCommandRetrieve *>( cmd );
	QList< int > results = GDSPnrDatabase::instance()->getByAID( c->id(), m_login );
	if ( results.length() == 0)
	{
		sendContent( "NOT FOUND" );
		return;
	}
	int i = 1;
	foreach( int p, results )
	{
		GDSPnr * f = GDSPnrDatabase::instance()->get( p );
		sendContent( QString( "%1 %2" ).arg( i ).arg( f->toPrettyString() ) );
		delete f;
		i++;
	}
}

void GDSClientAgent::processDeletePNR( GDSCommand* cmd )
{
	stateCheck( m_state != NORMAL );
	GDSCommandDeletePNR *c = dynamic_cast<GDSCommandDeletePNR *>( cmd );
	int n = 0;
	if ( !c->id().isEmpty() )
	{
		n = GDSPnrDatabase::instance()->deleteByAID( c->id(), m_login );
	}
	sendContent( QString( "Deleted %1 PNRs by id %2" ).arg( n ).arg( c->id() ) );
}

void GDSClientAgent::processChangePassword( GDSCommand* cmd )
{
	stateCheck( m_state == VOID );
	GDSCommandChangePassword* c = dynamic_cast< GDSCommandChangePassword* >( cmd );
	GDSClientAuthInfo* auth = GDSClientDatabase::instance()->getAuthByLogin( m_login );
	if ( c->password().isEmpty() || !auth || !auth->isChangeable() )
	{
		sendContent( "Can't change password" );
		return;
	}
	auth->setPassword( c->password() );
	sendContent( "Password changed" );
}
