#include <cryptlib.h>
#include <filters.h>
#include <rsa.h>
#include <sha.h>
#include <hex.h>
#include <base64.h>

#include "gdsclient_airline.h"
#include "gds_command.h"
#include "gdsclientsbase.h"
#include "gdsdatabase_flights.h"
#include "gdsdatabase_hotels.h"

using namespace CryptoPP;

bool GDSClientAirline::processContent( const QByteArray & data )
{
	m_log.proto( "Incoming content:", data );
	bool result = false;
	GDSCommand* cmd = GDSCommandCreator::create( data );
	try
	{
		if ( dynamic_cast<GDSCommandBatch *>( cmd ) )
		{
			result = processBatch( cmd );
		}
		else
		{
			result = addBatchCommand( data );
		}
	}
	catch (InternalError& e)
	{

	}
	delete cmd;
	return result;
};

bool GDSClientAirline::processAddFlight( GDSCommand* cmd )
{
	GDSCommandAddFlight* c = dynamic_cast<GDSCommandAddFlight*>( cmd );
	GDSFlightDatabase::instance()->add( GDSFlight( c ) );
	return true;
}

bool GDSClientAirline::processAddHotel( GDSCommand* cmd )
{
	GDSCommandAddHotel* c = dynamic_cast<GDSCommandAddHotel*>( cmd );
	GDSHotelDatabase::instance()->add( GDSHotel( c ) );
	return true;
}

bool GDSClientAirline::processAddUser( GDSCommand* cmd )
{
	GDSCommandAddUser *c = dynamic_cast< GDSCommandAddUser * >( cmd );
	GDSClientDatabase::instance()->addClient( c->login(), c->password(), c->changeable() );
	return true;
}

bool GDSClientAirline::processBatch( GDSCommand* c )
{
	GDSCommandBatch* cmd = dynamic_cast<GDSCommandBatch *>( c );
	m_signature = cmd->RSAsignature();
	m_commandCount = cmd->count();
	m_commands.clear();
	return true;
}

bool GDSClientAirline::addBatchCommand( const QByteArray& data )
{
	m_commands.push_back( data );
	m_commands.last().detach();
	if ( m_commands.size() < m_commandCount )
	{
		return true;
	}

	if ( checkSignature() )
	{
		m_log.info( "Batch message verified. Applying the changes" );
		foreach( const QByteArray& data, m_commands )
		{
			GDSCommand* cmd = GDSCommandCreator::create( data );
			if ( dynamic_cast<GDSCommandAddFlight *>( cmd ) )
			{
				processAddFlight( cmd );
			}
			else if ( dynamic_cast<GDSCommandAddHotel *>( cmd ) )
			{
				processAddHotel( cmd );
			}
			else if ( dynamic_cast<GDSCommandAddUser *>( cmd ) )
			{
				processAddUser( cmd );
			}
			delete cmd;
		}
	}
	else
	{
		m_log.warn( "Batch message bad signature" );
	}
	m_commandCount = 0;
	return true;
}

bool GDSClientAirline::checkSignature()
{
	StringSource pub( public_key, true, new Base64Decoder );
	RSASSA_PKCS1v15_SHA_Verifier pubVerifier( pub );

	StringSource signature( ( const byte * ) m_signature.toLocal8Bit().data(), m_signature.toLocal8Bit().size(), true, new HexDecoder );

	if ( signature.MaxRetrievable() != pubVerifier.SignatureLength() )
		return false;

	SecByteBlock byteBlock( pubVerifier.SignatureLength() );
	signature.Get( byteBlock, byteBlock.size() );


	VerifierFilter *verifierFilter = new VerifierFilter( pubVerifier );
	verifierFilter->Put( byteBlock, byteBlock.size() );

	QByteArray data;
	foreach( const QByteArray& cmd, m_commands )
	{
		data.append( cmd );
	}

	StringSource f( ( const byte * ) data.data(), data.size(), true, verifierFilter);

	return verifierFilter->GetLastResult();
}

GDSClientAirline::GDSClientAirline( GDSClientConnection * connection )
: GDSClient( connection )
{

}

GDSClientAirline::~GDSClientAirline()
{

}

// public RSA key
const char GDSClientAirline::public_key[]=
"MHwwDQYJKoZIhvcNAQEBBQADawAwaAJhALCUCIFzJL9qgBSL2pjAQBr/PZNWKJSg"
"a5rZQ+x0/zRXk5194ZKPCBqnzDgowa9sv0ljZEXstOcWeKXxKRoRtUAFlOwsxtDl"
"xg/iBEshK35/ll92Ic0ql6oAj6C8DrWceQIDAQAB";
