#include "gdsclientconnection.h"
#include "gdsclientauthenticator.h"
#include <QTcpSocket>
#include <QHostAddress>
#include "logstream.h"
#include "exception.h"

static const char packetSeparator = '\n';

GDSClientConnection::GDSClientConnection( QTcpSocket *socket )
: m_socket( socket )
, m_log( QString( "GDSClientConnection(%1:%2)" ).arg( socket->peerAddress().toString(), QString::number( socket->peerPort() ) ), LogStream::instance() )
{
	m_log.info( "Connection initiated" );
	check( connect( m_socket, SIGNAL( disconnected() ), this, SLOT( connectionClosedSlot() ) ) );
	check( connect( m_socket, SIGNAL( readyRead() ), this, SLOT( readDataSlot() ) ) );
	check( connect( m_socket, SIGNAL( error( QAbstractSocket::SocketError ) ), this, SLOT( errorSlot( QAbstractSocket::SocketError ) ) ) );
	new GDSClientAuthenticator( this );
}

void GDSClientConnection::readDataSlot()
{
	int dataSize = m_socket->bytesAvailable();
	int oldDataSize = m_readData.size();
	m_readData.resize( oldDataSize + dataSize );
	m_socket->read( m_readData.data() + oldDataSize, dataSize );

	int beg;
	int pos;
	for ( beg = 0; beg < m_readData.size() && ( pos = m_readData.indexOf( packetSeparator, beg ) ) != -1; beg = pos + 1 )
	{
		if ( pos )
		{
			QByteArray foundBlock = QByteArray::fromRawData( &( m_readData.data()[beg] ) , pos - beg );
			newPacket( foundBlock );
		}
	}

	if ( beg )
	{
		int tailSize = m_readData.size() - beg;
		if ( tailSize > 0 )
			memmove( m_readData.data(), m_readData.data() + beg, tailSize );
		m_readData.resize( tailSize );
	}
};

void GDSClientConnection::connectionClosedSlot()
{
	m_log.info( "Connection closed." );
	closeConnection();
}

void GDSClientConnection::closeConnection()
{
	m_socket->abort();
	m_socket->deleteLater();
	emit connectionClosedSignal( this );
	deleteLater();
}

void GDSClientConnection::sendContent( const QByteArray& content )
{
	m_log.proto( "Outgoing content:", content );
	m_socket->write( content );
}

void GDSClientConnection::connectPacketTarget( QObject* reciever, const char* slot )
{
	check( connect( this, SIGNAL( newPacketSignal( const QByteArray& ) ), reciever, slot) );
}

void GDSClientConnection::disconnectPacketTarget( QObject* reciever, const char* slot )
{
	check( disconnect( this, SIGNAL( newPacketSignal( const QByteArray& ) ), reciever, slot) );
}

void GDSClientConnection::connectDestroyedTarget( QObject* reciever, const char* slot )
{
	check( connect( this, SIGNAL( destroyed(QObject*) ), reciever, slot) );
}

void GDSClientConnection::disconnectDestroyedTarget(QObject* reciever, const char* slot )
{
	check( disconnect( this, SIGNAL( destroyed(QObject*) ), reciever, slot) );
}

void GDSClientConnection::registerUser( const QString& user )
{
	m_log.setDescription( QString( "GDSClientConnection (%1)" ).arg( user ) );
}

void GDSClientConnection::errorSlot( QAbstractSocket::SocketError error )
{
	if ( error != QAbstractSocket::RemoteHostClosedError )
	{
		m_log.error( QString( "Socket '0x%2' error: %3 (%4)." ).arg( QString::number( ( size_t ) &m_socket, 16 ), QString::number( error ), m_socket->errorString() ) );
	}
	closeConnection();
}

void GDSClientConnection::newPacket( const QByteArray& data )
{
	emit newPacketSignal( data );
}
